#
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $RCSfile: HADRSLA.pm,v $ $Revision: 1.1.2.9 $
#

package HADRSLA;

use ag_i18n_inc;
use Time::Local;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use POSIX qw(uname);
use Fcntl qw(:flock SEEK_END); # import LOCK_* and SEEK_END constants

use constant MODE_UNKNOWN => 0;
use constant MODE_READER => 1;
use constant MODE_WRITER => 2;

my $DEVNULL = "/dev/null";


sub rmsp {
    my $len = $#_ + 1;
    for (my $i = 0; $i<$len; $i++) {
        $_[$i] =~ s/^\s+//;
        $_[$i] =~ s/\s+$//;
    }
}

sub run_cmd {
	my ($cmd, $op) = @_;
	
	VCSAG_LOGDBG_MSG(3, "run_cmd: Issuing cmd: [$cmd]");
	my $typ = ref($op);
	if (!$typ) {
		system($cmd);
		return;
	}
	my $out = `$cmd`;
	my $ret = $?;
	rmsp $out;
	if ($typ eq "ARRAY")  {
		@{$op} = split(/\n/, $out);
	}
	if ($typ eq "SCALAR")  {
		${$op} = $out;
	}
	return $ret;
}

sub MakeRW {
    VCSAG_LOGDBG_MSG(4, "MakeRW : Begin");
    my $self = shift;
    my $rw = `$ag_i18n_inc::HACLUS -value ReadOnly`;
    if ( $rw == 1 ) {
        `$ag_i18n_inc::HACONF -makerw`;
    }
    VCSAG_LOGDBG_MSG(4, "MakeRW : End");
    return $rw;
}

sub MakeRO {
    VCSAG_LOGDBG_MSG(4, "MakeRO : Begin");
    my $self = shift;
    my $rw = `$ag_i18n_inc::HACLUS -value ReadOnly`;
    if ( $rw == 0 ) {
        `$ag_i18n_inc::HACONF -dump -makero`;
    }
    VCSAG_LOGDBG_MSG(4, "MakeRO : End");
}

sub CheckGCO {
	my $ResName = shift;
	my $GrpName = `$ag_i18n_inc::HARES -value $ResName Group`;
	chomp $GrpName;
	my $ClusterList = `$ag_i18n_inc::HAGRP -value $GrpName ClusterList`;
	chomp $ClusterList;
	my @Clust_array = split(/\s+/, $ClusterList);
	return 1 if ($#Clust_array == 3);
	return 0;
}

sub Dep_DG_Res {
	my (@lines, $line);
	my $ResName = shift;
	my $dep = `$ag_i18n_inc::HARES -dep $ResName`;
	@lines = split (/\n/, $dep);
	foreach $line (@lines){
		my @tokens = split (/\s+/, $line);
		if ($tokens[2] eq $ResName){
			my $ParentType = `$ag_i18n_inc::HARES -value $tokens[1] Type`;
			chomp $ParentType;
			if ($ParentType eq "DiskGroup") {
				return 1;
			}
		}
	}
	return 0;
}

sub VxVM_installed {
	my $ResName = shift;
	my (@lines, $dep, $dep_dg_res);
	my ($vxvm, $check_vxvm);
	$vxvm = 0;
	$check_vxvm = `vxdctl mode`;
	if ($? != -1){
		if ($check_vxvm =~ m/enabled/){
			$vxvm = 1;
			VCSAG_LOGDBG_MSG(2,"VXVM is installed and running.\n");
		}elsif ($check_vxvm =~ m/disabled/){
			$vxvm = 2;
			VCSAG_LOGDBG_MSG(2,"VxVM is installed but not running.\n");
		}
	}else{
		VCSAG_LOGDBG_MSG(2,"You need to install VxVM for computing Appliction RPO.\n");
	}

	$dep_dg_res = Dep_DG_Res($ResName);
	if ($vxvm == 1 && $dep_dg_res == 1){ # Good to Go
		return 1;
	}elsif ($vxvm == 0 ){ # No VxVM installed
		return 2;
	}elsif($vxvm == 1 && $dep_dg_res == 0){ # VxVm installed but DG resource not configured
		VCSAG_LOGDBG_MSG(2,"No DiskGroup resource dependent on the [$ResName] is defined. Cannot Compute Application RPO.\n");
		return 3;
	}elsif($vxvm == 2){ # VxVM is disabled can be enabled by using vxconfigd -k
		VCSAG_LOGDBG_MSG(2,"VxVM should be enabled for computing Application's RPO.\n");
		return 4;
	}
}




sub FindParentDGs {
	my ($ResName, $DGs) = @_;
    	my ($Group, $GroupDep, $DependentGroup, @resources);
	my @lines = ();
    	run_cmd("$ag_i18n_inc::HARES -dep $ResName", \@lines);
	foreach my $line (@lines) {#if the diskgroup resource and the srdf resource are in the same SG
		my @tokens = split(/\s+/, $line);
		if ($tokens[2] eq $ResName) {
			# Verify that the parent resource is of type DiskGroup.
			my $ParentType;
			run_cmd("$ag_i18n_inc::HARES -value $tokens[1] Type", \$ParentType);

			rmsp $ParentType;
			#VCSAG_LOGDBG_MSG(2, "FindParentDGs: Parent type is [$ParentType]");
			if ($ParentType eq "DiskGroup") {
				my $DiskGroup;
				run_cmd("$ag_i18n_inc::HARES -value $tokens[1] DiskGroup", \$DiskGroup);
				rmsp $DiskGroup;
				VCSAG_LOGDBG_MSG(2, "FindParentDGs: Parent DGName: [$DiskGroup]");
				push (@{$DGs}, $DiskGroup);
			}
		}
	}
}

sub FindVols {
	my ($DGs, $dg_disk_list_str, $dg_vol_assoc_str) = @_;

	my @lines = ();
	foreach my $DG (@{$DGs}) {
		next if ($DG =~ /^$/);
		VCSAG_LOGDBG_MSG(3, "FindVols: Fetching DG info for [$DG]");
		# Find volumes and disks from this DG
		@lines = ();
		run_cmd("vxprint -g $DG", \@lines);
		foreach my $line (@lines) {
			if ($line =~ /^v\s+(\S+)/) {
				my $vol = $1;
				#VCSAG_LOGDBG_MSG(2, "FindVols: Found volume [$vol] on DG [$DG]");
				$$dg_vol_assoc_str .= "$DG $vol ";
				next;
			}
			if ($line =~ /^dm\s+\S+\s+(\S+)/) {
				my $disk = $1;
				#VCSAG_LOGDBG_MSG(2, "FindVols: Found disk [$disk] in DG [$DG]");
				$$dg_disk_list_str .= "$disk ";
			}
		}
	}
	rmsp ($$dg_vol_assoc_str, $$dg_disk_list_str);

	VCSAG_LOGDBG_MSG(2, "FindVols: dg_vol_assoc_str: [$$dg_vol_assoc_str]");
	VCSAG_LOGDBG_MSG(2, "FindVols: dg_disk_list_str: [$$dg_disk_list_str]");
}

sub GetAttr {
	my ($ResName, $Attr) = @_;

	my $ComputeDRSLA;
	run_cmd("$ag_i18n_inc::HARES -value $ResName $Attr", \$ComputeDRSLA);
	rmsp $ComputeDRSLA;
	return $ComputeDRSLA;
}

sub SetAttr {
	my ($ResName, $Attr, $Val) = @_;

	run_cmd("$ag_i18n_inc::HACONF -makerw", 0);
	run_cmd("$ag_i18n_inc::HARES -modify $ResName $Attr $Val", 0);
	run_cmd("$ag_i18n_inc::HACONF -dump -makero", 0);
}

sub WriteTags {
	my $ResName = shift;
	
	# Read stats from file.
	my $filename = "$ResName" . "_DRSLA.txt";
	my @all_stats;
	my ($SessionId, $DRSLADGVols, $disklist, $disk_stats);
	# metadata1 contains SessionId No. metadata2 contains disk list.
	# TODO : Actions to perform if the file is missing
	if (-f $filename and !-z $filename) {
		open STATS, "+<", $filename;
		flock STATS, LOCK_EX;

		$SessionId = <STATS>;
		$DRSLADGVols = <STATS>;
		my $disk_n_stats = <STATS>;

		rmsp ($SessionId, $DRSLADGVols, $disk_n_stats);
		VCSAG_LOGDBG_MSG(2, "WriteTags: SessionId:DRSLADGVols [$SessionId:$DRSLADGVols]");
		@all_stats = <STATS>;
		#close STATS;
		if ($disk_n_stats =~ /^([^:]*):(.*)$/) {
			($disklist, $disk_stats) = ($1, $2);
			rmsp ($disklist, $disk_stats);
		} else {
			$disklist = $disk_n_stats;
		}
		VCSAG_LOGDBG_MSG(2, "WriteTags: Found disk list: [$disklist] and disk_stats: [$disk_stats]");
	}
	my %stats = {};
	foreach my $line (@all_stats) {
		rmsp $line;
		next if $line =~ /^$/;
		$line =~ /^(\S+)\s+(\S+)\s+(.*)$/;
		my ($dg, $vol, $all_writes) = ($1, $2, $3);
		$stats{$dg}{$vol}{'all_writes'} = $all_writes;
		$line =~ /(\S+)$/;
		$stats{$dg}{$vol}{'last_writes'} = $1;
		VCSAG_LOGDBG_MSG(2, "WriteTags: LastWrites on DG:Vol: [$dg:$vol]: [$stats{$dg}{$vol}{'all_writes'}], [$stats{$dg}{$vol}{'last_writes'}]");
	}
	# For the first time this part of the code will be run.
	my @token = split (/\s+/, $DRSLADGVols);
	my $written = 0;
	my (@out_stats, @lines);
	for (my $i = 0; $i < $#token + 1; $i+=2) {
		my ($dg, $vol) = ($token[$i], $token[$i+1]);
		VCSAG_LOGDBG_MSG(2, "WriteTags: Fetching current stats for DG [$dg] Volume: [$vol]");
		@lines = ();
		run_cmd("vxstat -g $dg $vol", \@lines);
		my @vol_lines = grep(/^vol/, @lines);
		foreach my $line (@vol_lines) {
			my @stats = split(/\s+/, $line);
			my $writes = $stats[3];
			$stats{$dg}{$vol}{'curr_writes'} = $writes;
			VCSAG_LOGDBG_MSG(2, "WriteTags: Writes on DG:Vol [$dg:$vol]: [$writes]");
                        if (($stats{$dg}{$vol}{'last_writes'} != $stats{$dg}{$vol}{'curr_writes'}) || (!$stats{$dg}{$vol}{'curr_writes'})) {
				VCSAG_LOGDBG_MSG(2, "WriteTags: Write counter has incremented from [$stats{$dg}{$vol}{'last_writes'}]");
				$written = 1;
			}
			my $out_stat = "$dg $vol $stats{$dg}{$vol}{'curr_writes'}";
			push @out_stats, $out_stat;
		}
	}
	my $TagName = "RPOTag";

	my $SeqNo = time;
	VCSAG_LOGDBG_MSG(2, "WriteTags: disk_stats: [$disk_stats]");
	$disk_stats =~ s/(\S+)$//;
	my $last_tag = $1;
	my $old_tags = $disk_stats;
	rmsp($old_tags, $last_tag);
	VCSAG_LOGDBG_MSG(2, "WriteTags: old_tags: [$old_tags], last_tag: [$last_tag]");
	my $new_taglist;
	if ($written) {
		# Write the tag on the disk.
		my $DiskTag = $SessionId . "_" . $SeqNo;
		VCSAG_LOGDBG_MSG(2, "WriteTags: SessionId:SeqNo [$SessionId:$SeqNo]");
		VCSAG_LOGDBG_MSG(2, "WriteTags: new DiskTag: [$DiskTag]");

		my @disks = split(/\s+/, $disklist);
		foreach my $disk (@disks) {
			run_cmd("vxdisk rmtag $TagName $disk > $DEVNULL 2>&1", 0);
			run_cmd("vxdisk settag $TagName=$DiskTag $disk", 0);
		}
		$new_taglist = "$old_tags $last_tag $SeqNo:$SeqNo";
	} else {
		$last_tag =~ /(\d+):(\d+)/;
		my ($old_SeqNo_from, $old_SeqNo_to) = ($1, $2);
		$new_taglist = "$old_tags $old_SeqNo_from:$SeqNo";
	}
	rmsp($new_taglist);
	VCSAG_LOGDBG_MSG(2, "WriteTags: new_taglist: [$new_taglist]");
	seek STATS, 0, 0;
	truncate STATS, 0;
	print STATS "$SessionId\n";
	print STATS "$DRSLADGVols\n";
	print STATS "$disklist:$new_taglist\n";
	foreach my $line (@out_stats) {
		print STATS "$line\n";
	}
	close STATS;
}

sub FindRemoteClusNode
{
	my ($ResName, $localmode) = @_;
	my ($RemoteClus, $RemoteNode);

	my @lines = ();
	VCSAG_LOGDBG_MSG(2, "FindRemoteClusNode: Entered with mode [$localmode].");
	if ($localmode eq MODE_READER) { # Reader mode.
		run_cmd("$ag_i18n_inc::HARES -state $ResName", \@lines);
		foreach my $line (@lines) {
			if ($line =~ /(\S+):(\S+)\s+ONLINE$/) {
				($RemoteClus, $RemoteNode) = ($1, $2);
				last;
			}
		}
	} elsif ($localmode eq MODE_WRITER) { # Writer mode.
		my ($Group, $ClusterList, $LocalClus, @SysStates);
		run_cmd("$ag_i18n_inc::HARES -value $ResName Group", \$Group);
		run_cmd("$ag_i18n_inc::HAGRP -value $Group ClusterList", \$ClusterList);
		run_cmd("$ag_i18n_inc::HACLUS -value ClusterName", \$LocalClus);

		# TODO: Add support for multiple clusters.
		my @ClusList = split(/\s+/, $ClusterList);
		my $DRClus;
		for (my $i=0; $i < $#ClusList + 1; $i+=2) {
			if ($ClusList[$i] ne $LocalClus) {
				$DRClus = $ClusList[$i];
			}
		}

		run_cmd("$ag_i18n_inc::HAGRP -value $Group SystemList -clus $DRClus", \@lines);
		$lines[1] =~ /^\S+\s+\S+\s+\S+\s+(.*)/;
		my $DRSystemList = $1;
		my @DRSysList = split(/\s+/, $DRSystemList);

		run_cmd("$ag_i18n_inc::HASYS -state", \@SysStates);
		my @RunningSys = grep(/RUNNING/, @SysStates);
		my @RunningDRSys = grep(!/localclus/, @RunningSys);

		my $DRNode;
		for (my $i=0; $i < $#DRSysList + 1; $i+=2) {
			my @RunningDRSysList = grep (/$DRSysList[$i]/, @RunningDRSys);
			if ($#RunningDRSysList +1 >= 0) {
				$DRNode = $DRSysList[$i];
			}#TODO : Will it always online on the last system in the systemlist. 
		}
        $RemoteClus = $DRClus;
        $RemoteNode = $DRNode;
	} else {
		VCSAG_LOGDBG_MSG(2, "FindRemoteClusNode: Unknown localmode. Could not find remote node.");
		return;
	}

	VCSAG_LOGDBG_MSG(2, "FindRemoteClusNode: returning Clus:Node: [$RemoteClus:$RemoteNode]");
	return ($RemoteClus, $RemoteNode);
}

sub ReadTags {
	my $ResName = shift;

	my $filename = $ResName . "_DRSLA.txt";
	open DISKS, $filename;
	flock DISKS, LOCK_SH;
	my $SessionId = <DISKS>;
	my $Disks = <DISKS>;
	close DISKS;
	rmsp($SessionId, $Disks);

	VCSAG_LOGDBG_MSG(2, "ReadTags: Expected SessionId: [$SessionId]");

	my $Tagname = "RPOTag";
	# Find the smallest sequence no
	my @lines = ();
	run_cmd("vxdisk listtag $Disks |grep $Tagname|sort -k3,3 -u", \@lines);
	$lines[0] =~ /(\d+)_(\d+)$/;
	my ($SessId, $SeqNo) = ($1, $2);

	VCSAG_LOGDBG_MSG(2, "ReadTags: Read SessionId: [$SessId]");
	if ($SessId ne $SessionId) {
		VCSAG_LOGDBG_MSG(2, "ReadTags: SessionId mismatch. Not reporting any data.");
		return;
	}
		

	# Find the node in the primary cluster where the Replication resource is online.
	# Then report this data to that node for RPO computation.
	
	my ($PrimClus, $PrimNode) = FindRemoteClusNode($ResName, MODE_READER);
	
    if ($PrimClus && $PrimNode){
        @lines = ();
        run_cmd("$ag_i18n_inc::HARES -action $ResName ReportRPOData -actionargs $SeqNo -sys $PrimNode -clus $PrimClus", \@lines);
    } else {
        VCSAG_LOGDBG_MSG( 2," Remotecluster and Remote system not found");
        return;
    }
    my $CurrRPO = 0;
	foreach (@lines) {
		if (/Current RPO:\s+(\d+)$/) {
			$CurrRPO = $1;
			VCSAG_LOGDBG_MSG(2, "ReadTags: Current RPO: [$CurrRPO]");
		}
	}
	my $curr_time = time;
	open DISKS, ">>", $filename;
	flock DISKS, LOCK_EX;
	print DISKS "$curr_time $CurrRPO\n";
	close DISKS;
	#SetAttr($ResName, "CurrentRPO", $CurrRPO);
}

# Finds whether we should run in reader (1) or writer (2) mode. (0) if unknown.
sub FindMode {
	my $ResName = shift;

	my ($line, @lines);
	run_cmd("$ag_i18n_inc::HASYS -nodeid", \$line);
	$line =~ /Node name:\s+(\S+)/;
	my $nodename = $1;
	run_cmd("$ag_i18n_inc::HARES -state $ResName ", \@lines);
	my @online_any = grep (/ONLINE/, @lines);
	my $num_lines = $#online_any;
	if ($num_lines == -1) {
		VCSAG_LOGDBG_MSG(2, "FindMode:Resource is not online in any of the cluster. Cannot compute RPO."); 
		return MODE_UNKNOWN;
	}

    my @local_nodes = grep (/localclus.*ONLINE/, @lines);
    if ( $#local_nodes == -1 ){
        VCSAG_LOGDBG_MSG(2, " FindMode : Resource is not online in localclus Reader Mode.");
        return MODE_READER;
    }

	foreach my $line (@online_any) {
		rmsp $line;
		if ($line =~ /$nodename/) {
			# This is the writer node.
			VCSAG_LOGDBG_MSG(2, "FindMode: This is the writer node.");
			#WriteTags($ResName);
			VCSAG_LOGDBG_MSG(2, "FindMode: Localmode: MODE_WRITER");
			return MODE_WRITER;
		}
	}
	VCSAG_LOGDBG_MSG(2, "FindMode: Localmode: MODE_UNKNOWN");
	return MODE_UNKNOWN;
}

sub DGResOnline {
    my $ResName = shift;
    my (@lines, $state);
    run_cmd("$ag_i18n_inc::HARES -dep $ResName", \@lines);
    foreach my $line (@lines){
        my @tokens = split (/\s+/, $line);
        if ($tokens[2] eq $ResName){
            my $ParentType;
            run_cmd("$ag_i18n_inc::HARES -value $tokens[1] Type", \$ParentType);
            rmsp $ParentType;
            if ($ParentType eq "DiskGroup") {
                run_cmd("$ag_i18n_inc::HARES -value $tokens[1] State", \$state);
                rmsp $state;
                return 1 if ($state eq "ONLINE");
            }
        }
    }
    return 0;
}

sub MainLoop {
    my ($ResName, $computedrsla) = @_;
    my $filename = "$ResName" . "_DRSLA.txt";

    if ($computedrsla == 0 || $computedrsla == 1){
    }else{
        VCSAG_LOGDBG_MSG(2, "MainLoop: Value of computeDRSLA attribute should be 0 or 1. Current Valus is [$computedrsla]");
        return;
    }
    
	#VCSAG_LOGDBG_MSG(2, "MainLoop: Checking whether to compute RPO.");
    if (-f $filename && FindMode($ResName) == MODE_WRITER && DGResOnline($ResName)) {
        WriteTags($ResName);
    } elsif (!-f $filename && FindMode($ResName) == MODE_WRITER && DGResOnline($ResName)){
        my ($tagging) = split(/\n/, `$ag_i18n_inc::HARES -value $ResName Tagging`);
        if ($tagging){
            StartWriter($ResName);
        }
    } else {
        VCSAG_LOGDBG_MSG(2, "MainLoop: Could not find the mode. Not doing anything in this iteration.")
    }
}

sub StartWriter {
	# TODO: Accept identity of reader so that we can maintain multiple secondaries.
	my $ResName = shift;

	clean_files($ResName);

	# Find the VxVM DiskGroup depending on ResName
	my (@DGs, $dg_disk_list_str, $dg_vol_assoc_str) = ((), "", "");
	FindParentDGs($ResName, \@DGs);
	FindVols(\@DGs, \$dg_disk_list_str, \$dg_vol_assoc_str);

	my $filename = "$ResName" . "_DRSLA.txt";
    my $rw = MakeRW();
    `$ag_i18n_inc::HARES -modify $ResName Tagging 1`;
    if ($rw){
        MakeRO();
    }
    srand(time);
	my $RandInt = int (rand(1000000));
	my $SessionId = sprintf("%06d", $RandInt);

	open STATS, ">", $filename;
	flock STATS, LOCK_EX;
	print STATS "$SessionId\n";
	print STATS "$dg_vol_assoc_str\n";
	print STATS "$dg_disk_list_str\n";
	close STATS;
	WriteTags($ResName);
	print "SessionId: $SessionId\n";
}

sub StartRPOComputation {
	my $ResName = shift;
	my $filename = $ResName . "_DRSLA.txt";
	my $SessId;

	VCSAG_LOGDBG_MSG(2, "StartRPOComputation Entered.");

	my $mode = FindMode($ResName);
    if (!$mode){
        return;
    }
    VCSAG_LOGDBG_MSG(2, "StartRPOComputation: Mode = [$mode]");
	if ($mode ne MODE_READER) {
		VCSAG_LOGDBG_MSG(2, "StartRPOComputation: This action should be performed only on the secondary site.");
		return;
	}
	# Find the first RUNNING node in the DR SG's SystemList and send a message (action EP) to start reading Tags.
	my ($PriClus, $PriNode) = FindRemoteClusNode($ResName, MODE_READER);

	my (@lines, @vxdisk_list);
    if ($mode eq MODE_READER && !(-f $filename)) {
        if ($PriClus && $PriNode){
            run_cmd("$ag_i18n_inc::HARES -action $ResName StartWriter -sys $PriNode -clus $PriClus", \@lines);
        } else {
            VCSAG_LOGDBG_MSG( 2," Remotecluster and Remote system not found");
            return;
        }
        open DISKS, ">", $filename;
	    foreach (@lines) {
		    rmsp $_;
		    if (/^SessionId:\s+(\d+)$/) {
			    $SessId = $1;
			    last;
		    }
	    }
    }
	VCSAG_LOGDBG_MSG(2, "StartRPOComputation: Started Writer with SessionID: [$SessId]");

	# Find DiskGroup depending on ResName
	# TODO : If the DiskGroup name are DG1, DG12, DG13 then the grep(/$DG_pat/, @vxdisk_list)
	# will not work we must check if the DG name is eq to the diskgroup mentioned in thr DG resource
	# or use delimiter in the Grep
	my @DGs = ();
	FindParentDGs($ResName, \@DGs);
	my $DG_pat = join('|', @DGs);
	VCSAG_LOGDBG_MSG(2, "StartRPOComputation: Search Pattern: [$DG_pat]");
	run_cmd("vxdisk -o alldgs list", \@vxdisk_list);
	my @dg_disklist = grep(/$DG_pat/, @vxdisk_list);

	# TODO: Check for empty array.

	my $RepDisks;
	foreach my $line (@dg_disklist) {
		$line =~ /^(\S+)/;
		$RepDisks .= "$1 ";
	}
	chop($RepDisks);

	flock DISKS, LOCK_EX;
	print DISKS "$SessId\n";
	print DISKS "$RepDisks\n";
	close DISKS;

}

sub clean_files {
	my $ResName = shift;
	my $filename = "$ResName" . "_DRSLA.txt";
	unlink $filename;
}


sub StopRPOComputation {
	my $ResName = shift;

	VCSAG_LOGDBG_MSG(2, "StopRPOComputation Entered.");

	clean_files($ResName);

	my $mode = FindMode($ResName);
	if ($mode eq MODE_READER) {
	    VCSAG_LOGDBG_MSG(2, "StopRPOComputation: Sending Stop message to Primary node.");
        my ($PriClus, $PriNode) = FindRemoteClusNode($ResName, MODE_READER);
        if ($PriClus && $PriNode){
	        run_cmd("$ag_i18n_inc::HARES -action $ResName StopRPOComputation -sys $PriNode -clus $PriClus", 0);
        } else {
            VCSAG_LOGDBG_MSG( 2," Remotecluster and Remote system not found");
            return;
        }
	    VCSAG_LOGDBG_MSG(2, "StopRPOComputation: Reader stopped");
	    return 0;
    }
    if ($mode eq MODE_WRITER){
        my $rw = MakeRW();
        `$ag_i18n_inc::HARES -modify $ResName Tagging 0`;
        if ($rw){
            MakeRO();
        }
    }
}

sub ReportRPOData {
	my ($ResName, $Sessid, $SeqNo) = @_;

	my $filename = "$ResName" . "_DRSLA.txt";
	my ($SessionId, $DRSLADGVols);
	open STATS, "+<", $filename;
	flock STATS, LOCK_EX;

	$SessionId = <STATS>;
	$DRSLADGVols = <STATS>;
	my $disk_n_stats = <STATS>;

	rmsp ($SessionId, $DRSLADGVols, $disk_n_stats);
	my @all_stats = <STATS>;

	my ($disklist, $disk_stats);
	if ($disk_n_stats =~ /^([^:]*):(.*)$/) {
		($disklist, $disk_stats) = ($1, $2);
	} else {
		$disklist = $disk_n_stats;
	}
	rmsp $disklist, $disk_stats;
	VCSAG_LOGDBG_MSG(2, "ReportRPOData: Received Session ID: [$Sessid], Received SeqNo: [$SeqNo], Found disk list: [$disklist] and disk_stats: [$disk_stats]");
	my $curr_rpo;

	if ($SessionId ne $Sessid) {
		VCSAG_LOGDBG_MSG (2, " Received Session ID: [$Sessid] and Current Session Id [$SessionId] didn't matched.");
		print "Session Id Not Matched\n";
		return ;
	}
	if ($disk_stats =~ /^.*($SeqNo:(\d+).*)/) {
		my $list_start = $1;
		my $end = $2;

		my ($CurrTime, $EndTime);
		if ($list_start =~ /^*?(\d+):(\d+)$/){
			$CurrTime = $1;
			$EndTime = $2;
		}
		
		$CurrTime == $SeqNo?($curr_rpo = 0):($curr_rpo = $CurrTime - $end);
		
		rmsp $disk_stats;
		VCSAG_LOGDBG_MSG(2, "ReportRPOData: CurrTime: [$CurrTime], [$SeqNo], end: [$end]");
		VCSAG_LOGDBG_MSG(2, "ReportRPOData: New diskstats: [$disk_stats]");

		seek STATS, 0, 0;
		truncate STATS, 0;
		print STATS "$SessionId\n";
		print STATS "$DRSLADGVols\n";
		print STATS "$disklist:$list_start\n";
		foreach my $line (@all_stats) {
			print STATS "$line";
		}
		close STATS;
		return "$curr_rpo";
	} else  {
		VCSAG_LOGDBG_MSG(2, "ReportRPOData: Invalid SeqNo received [$SeqNo].");
	}
}

sub time_format {
	my $time = shift;
	my ($hour, $min, $sec,$t1);
	$hour = "0";
	$min = "0";
	$sec = "0";
	if ($time >= 60){
		$t1 = int ($time / 60);
		$sec = ($time % 60);
		if ($t1 >= 60){
			$hour = int ($t1/60);
			$min = $t1 % 60;
		}else{
			$min = ($t1);
		}
	}
	else {
		$sec = $time;
	}
	$hour = "0$hour" if ($hour < 10 && $hour >= 0);
	$min = "0$min" if ($min < 10 && $min >= 0);
	$sec = "0$sec" if ($sec < 10 && $sec >= 0);
	return "$hour:$min:$sec";
}


sub GetCurrentRPO {
	my ($ResName, $computedrsla) = @_;

	my $filename = $ResName . "_DRSLA.txt";
	open DISKS, $filename;
	flock DISKS, LOCK_SH;
	my $SessionId = <DISKS>;
	my $Disks = <DISKS>;
	close DISKS;
	my ($SessId, $SeqNo);
	rmsp($SessionId, $Disks);
	my $mode = FindMode($ResName);
	if ($mode != MODE_READER){
		VCSAG_LOGDBG_MSG(2,"GetCurrentRPO: This action entry point should be invoked at the DR site.");
		print "This action entry point should be invoked at the DR site.";
		return 1;
	}
	if ($computedrsla != 1){
		VCSAG_LOGDBG_MSG (2, "GetCurrentRPO: Value of ComputeDRSLA should be 1 for computing RPO.");
		print "Value of ComputeDRSLA should be 1 for computing RPO.";
		return 1;
	}
	VCSAG_LOGDBG_MSG(2, "GetCurrentRPO: Expected SessionId: [$SessionId]");

	my $Tagname = "RPOTag";
# Find the smallest sequence no
	my @lines = ();
	run_cmd("vxdisk listtag $Disks |grep $Tagname|sort -k3,3 -u", \@lines);
	if ($#lines != -1){
          $lines[0] =~ /(\d+)_(\d+)$/;
          ($SessId, $SeqNo) = ($1, $2);
      } else {
          VCSAG_LOGDBG_MSG(2,"GetCurrentRPO: Cannot compute RPO. Wait for sometime and then fire GetCurrentRPO action again.\n");
      	  print "GetCurrentRPO: Cannot compute RPO. Wait for sometime and then fire GetCurrentRPO action again";
      	  return 1;
      }
	VCSAG_LOGDBG_MSG(2, "GetCurrentRPO: Read SessionId: [$SessId]");

# Find the node in the primary cluster where the Replication resource is online.
# Then report this data to that node for RPO computation.

	my ($PrimClus, $PrimNode) = FindRemoteClusNode($ResName, MODE_READER);
	if ($PrimClus && $PrimNode){
		@lines = ();
		run_cmd("$ag_i18n_inc::HARES -action $ResName ReportRPOData -actionargs $SessId $SeqNo -sys $PrimNode -clus $PrimClus", \@lines);
	} else {
		VCSAG_LOGDBG_MSG( 2," Remotecluster and Remote system not found");
		print "Remotecluster and Remote system not found";
		return 1;
	}
	my $CurrRPO = 0;
	my $Rep_Tech_RPO = "";
	if (@lines) {
		foreach (@lines) {
			if (/Current RPO:\s+(\d+)$/) {
				$CurrRPO = $1;
				VCSAG_LOGDBG_MSG(2, "ReadTags: Current RPO: [$CurrRPO]");
			}
			if (/Rep_Tech_RPO:\s+(.*)$/) {
				if ( $Rep_Tech_RPO eq "" ) {
					$Rep_Tech_RPO = $1;
				} else {
					$Rep_Tech_RPO = $Rep_Tech_RPO . "\n$1";
				}
				VCSAG_LOGDBG_MSG(2, "ReadTags: $Rep_Tech_RPO");
			}
			if (/Not Matched/) {
				print "\nNot able to compute RPO. Sessions are not matching.\n";
				return 1;
			}
		}
		my $C_RPO = time_format ($CurrRPO);
		print "\nApplication's RPO = $C_RPO  [hh:mm:ss]\n\n$Rep_Tech_RPO\n";
		VCSAG_LOG_MSG ("I", "Current RPO is [$CurrRPO]. $Rep_Tech_RPO\n", 1005, "$CurrRPO", "$Rep_Tech_RPO");
	return 0;
	} else {
		print "Not able to compute RPO.";
		VCSAG_LOG_MSG ("I", "Not able to compute RPO session id not matched.");
	}
}
1;

